/*  fitsfrompg                                                    */
/*  Jure Skvarc, 1998                                             */
/*  A program to find a FITS image which has header information   */
/*  stored in a POSTGRES database and  which satisfies            */
/*  certain criteria such as exposure time within limits,         */
/*  and which calculates image coordinates for the given celestial*/
/*  coordinates                                                   */
/*================================================================*/
#ifdef HPUX
#define _INCLUDE_HPUX_SOURCE
#define _INCLUDE_XOPEN_SOURCE
#define _INCLUDE_POSIX_SOURCE
#endif
#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <string.h>
#include <getopt.h>
#include <time.h>
#include <unistd.h>

#include <fitsio.h>
#include <libpq-fe.h>
#include "structs.h"
#include "consts.h"
#include "functs.h"

STATE state;

/*  Check if the required keyword was successfully read and exit otherwise  */
void
check_status_strong(char *key, int status)

{
  if (status != 0) {
    fprintf(state.logfile, "Keyword %s is missing. Can't continue!\n", key);
    exit(1);
  }
}

/*  Strip ' from the beginning and end of string, if it exists  */
void
strip(char *string) 

{
  if (string[0] == '\'') {
    strcpy(string, string + 1);
  }
  if (string[strlen(string) - 1] == '\'') {
    string[strlen(string) - 1] = '\0';
  }
}

/*  Divide the input parameter into filename only and the complete path to */
/*  the filename    */
int
resolve_filename(char *fname, char **path, char **name)

{
  char *p, t, *crd;
  
  crd = getcwd(NULL, 0);    /*  Get the current directory  */
  /*  Check if a directory is specified is the file name */
  p = strrchr(fname, '/');
  if (p == 0) {
    /*  There is only a file name  */
    *name = strdup(fname);
  }
  else {
    *name = strdup(p + 1);
    t = p[1];
    p[1] = '\0';
    /*  Change the directory  */
    if (chdir(fname)) {
      fprintf(stderr, "Can't chenge to directory %s\n", fname);
      return 1;
    }
    p[1] = t;
  }
  /*  Get the current directory  */
  *path = getcwd(NULL, 0);
  if (*path == 0) {
      fprintf(stderr, "Can't chenge to directory %s\n", fname);
      return 1;
  }
  /*  Chdir back to the original directory  */
  chdir(crd);
  free(crd);
  return 0;
}


int
main(int argc, char *argv[])
     
{
  int opt;
  int i, n;
  char a[2000];
  double xpix, ypix, ra, dec;
  int err;
  int inaxis1, inaxis2;
  int icrval1, icrpix1, icdelt1;
  int icrval2, icrpix2, icdelt2, icrot;
  int iproj, iarch, ipath, ifile;
  double offscreen = 0;
  int naxis1, naxis2;
  double crval1, crpix1, cdelt1;
  double crval2, crpix2, cdelt2, crot;
  char proj[5];
  char *archive = NULL, *path = NULL, *filename = NULL;
  int help = 0;

  PGconn* pgconn;
  PGresult* pgres;
  char *cra = NULL, *cdec = NULL;
  char *firstdate = NULL, *lastdate = NULL;
  char *other = NULL;

  state.server = NULL;
  state.database = "images";
  state.port = NULL;
  state.logfile = stdout;
  state.logname = NULL;
  state.version = "1.0";
  state.verbose = 0;
  /*  Check for command line options  */
  while ((opt = getopt (argc, argv, "l:s:d:p:vm:e:c:R:D:F:L:W:h")) != -1) {
    switch(opt) {
    case 'l':
      /*  Name of the log file  */
      state.logname = strdup(optarg);
      if ((state.logfile = fopen(state.logname,"a") )== NULL) {
	fprintf(stderr, "Can not open log file!\n");
	exit(1);
      }
      break;
    case 's':
      /*  Internet address of the server  */
      state.server = strdup(optarg);
      break;
    case 'd':
      /*  Name of the database  */
      state.database = strdup(optarg);
      break;
    case 'p':
      /*  Port number of the server  */
      state.port = strdup(optarg);
      break;
    case 'v':
      /*  Switch on verbose mode  */
      state.verbose = 1;
      break;
    case 'R':
      /*  right ascension  */
      cra = strdup(optarg);
      break;
    case 'D':
      /*  declination  */
      cdec = strdup(optarg);
      break;
    case 'F':
      /*  first date  */
      firstdate = strdup(optarg);
      break;
    case 'L':
      /*  last date  */
      lastdate = strdup(optarg);
      break;
    case 'W':
      /*  last date  */
      other = strdup(optarg);
      break;
    case 'E':
      /*  Margin of error for still acceptable coordinates  */
      offscreen = strtod(optarg, NULL);
      break;
    case 'h':
      help = 1;
      break;
    case '?':
      switch(optopt) {
      default:
	fprintf(stderr, "Unknown option -%c ignored!\n", optopt);
	break;
      }
      break;
    }
  }
  if (argc < optind) {
    printf("fits2pg %s     Jure Skvarc, September 1998\n", state.version);
    printf("Usage:: %s [options]\n", argv[0]);
    printf("Options:\n");
    printf("-l filename    name of the log file (default stdout)\n");
    printf("-s server      Internet address of the server\n");
    printf("-d database    database name\n");
    printf("-p port        port number\n");
    printf("-a archive     archive name\n");
    printf("-e email       email address\n");
    printf("-R ra          right ascension\n");
    printf("-D dec         declination\n");
    printf("-F date        first date\n");
    printf("-L date        last date\n");
    printf("-W query       additional conditions\n");
    printf("-h             print this message\n");
    printf("-v             verbose mode\n");
    printf("\n");
    exit(1);
  }
  
  /*===================================*/
  /* make a connection to the database */
  /*===================================*/
  pgconn = PQsetdb(state.server, state.port, NULL, NULL, state.database);
  
  /* check to see that the backend connection was successfully made */
  if (PQstatus(pgconn) == CONNECTION_BAD) {
    fprintf(stderr,"Connection to database '%s' failed!", state.database);
    fprintf(stderr,"%s",PQerrorMessage(pgconn));
    PQfinish(pgconn);
    exit(1);
  }
  if (state.verbose) {
    printf("Successfully connected to database %s on server %s\n", 
	   PQdb(pgconn), PQhost(pgconn));
    printf("The port used is %s.\n", PQport(pgconn));
  }

  printf("ra=%s, dec=%s, first=%s, last=%s, query=%s\n", cra, cdec, firstdate,
	 lastdate, other);

  /*=========================================================*/
  /*  Select the images which satisfy conditions imposed by  */
  /*  firstdate, lastdate and other                          */
  /*=========================================================*/
  sprintf(a, "SELECT * from images %s %s %s%s%s %s %s %s%s%s %s %s", 
	  (firstdate || lastdate || other) ? "WHERE" : "",
	  (firstdate) ? "date_mid >=" : "", (firstdate) ? "'" : "",
	  (firstdate) ? firstdate : "", (firstdate) ? "'" : "",
	  (firstdate && lastdate) ? "AND" : "", 
	  (lastdate) ? "date_mid <=" : "", (lastdate) ? "'" : "",
	  (lastdate) ? lastdate : "", (lastdate) ? "'" : "",
	  ((firstdate || lastdate) && other) ? "AND" : "", 
	  (other) ? other : "");
  if (state.verbose) {
    printf("Query=%s\n", a);
  }
  pgres = PQexec(pgconn, a);
  if (PQresultStatus(pgres) != PGRES_TUPLES_OK) {
    fprintf(state.logfile, "Error in making a query\n");
    PQclear(pgres);
    PQfinish(pgconn);
    exit(1);
  }
  if (state.verbose) {
    printf("Number of records:%d\n", PQntuples(pgres));
  }
  /*  For each image calculate where does the ra, dec coordinate fall  */
  /*  (if they are specified)                                          */
  if (cra && cdec) {
    ra = parse_ra_c(cra);
    dec = parse_dec_c(cdec);
    printf("ra=%f, dec=%f\n", ra, dec);
    /*===========================================================*/
    /*  Find out indices of keywords needed for wcs calculation  */
    /*===========================================================*/
    if ((inaxis1 = PQfnumber(pgres, "naxis1")) < 0) {
      fprintf(stderr, "naxis1 keyword not present in the database\n");
      goto out;
    }
    if ((inaxis2 = PQfnumber(pgres, "naxis2")) < 0) {
      fprintf(stderr, "naxis2 keyword not present in the database\n");
      goto out;
    }
    if ((icrval1 = PQfnumber(pgres, "crval1")) < 0) {
      fprintf(stderr, "crval1 keyword not present in the database\n");
      goto out;
    }
    if ((icrpix1 = PQfnumber(pgres, "crpix1")) < 0) {
      fprintf(stderr, "crpix1 keyword not present in the database\n");
      goto out;
    }
    if ((icdelt1 = PQfnumber(pgres, "cdelt1")) < 0) {
      fprintf(stderr, "cdelt1 keyword not present in the database\n");
      goto out;
    }
    if ((icrval2 = PQfnumber(pgres, "crval2")) < 0) {
      fprintf(stderr, "crval2 keyword not present in the database\n");
      goto out;
    }
    if ((icrpix2 = PQfnumber(pgres, "crpix2")) < 0) {
      fprintf(stderr, "crpix2 keyword not present in the database\n");
      goto out;
    }
    if ((icdelt2 = PQfnumber(pgres, "cdelt2")) < 0) {
      fprintf(stderr, "cdelt2 keyword not present in the database\n");
      goto out;
    }
    if ((icrot = PQfnumber(pgres, "crot")) < 0) {
      fprintf(stderr, "crot keyword not present in the database\n");
      goto out;
    }
    if ((iproj = PQfnumber(pgres, "projection")) < 0) {
      fprintf(stderr, "projection keyword not present in the database\n");
      goto out;
    }
    if ((iarch = PQfnumber(pgres, "archivename")) < 0) {
      fprintf(stderr, "archivename keyword not present in the database\n");
      goto out;
    }
    if ((ipath = PQfnumber(pgres, "path")) < 0) {
      fprintf(stderr, "path keyword not present in the database\n");
      goto out;
    }
    if ((ifile = PQfnumber(pgres, "filename")) < 0) {
      fprintf(stderr, "filename keyword not present in the database\n");
      goto out;
    }
    n = PQntuples(pgres);
    /*==========================*/
    /*  Loop over all records   */
    /*==========================*/
    err = 0;
    for (i = 0; i < n; i++) {
      crval1 = strtod(PQgetvalue(pgres, i, icrval1), NULL);
      crpix1 = strtod(PQgetvalue(pgres, i, icrpix1), NULL);
      cdelt1 = strtod(PQgetvalue(pgres, i, icdelt1), NULL);
      crval2 = strtod(PQgetvalue(pgres, i, icrval2), NULL);
      crpix2 = strtod(PQgetvalue(pgres, i, icrpix2), NULL);
      cdelt2 = strtod(PQgetvalue(pgres, i, icdelt2), NULL);
      crot = strtod(PQgetvalue(pgres, i, icrot), NULL);
      strncpy(proj, PQgetvalue(pgres, i, iproj), 5);
      fits_world_to_pix(ra, dec, crval1, crval2, crpix1, crpix2, 
			cdelt1, cdelt2, crot, proj, &xpix, &ypix, &err);
      if (!err) {
	double xmin, xmax, ymin, ymax;
	xpix--;
	ypix--;
	/*  Calculate limiting coordinates */
	/*=================================*/
	naxis1 = strtol(PQgetvalue(pgres, i, inaxis1), NULL, 10);
	naxis2 = strtol(PQgetvalue(pgres, i, inaxis2), NULL, 10);
	xmin = - naxis1 * offscreen;
	xmax = naxis1 * (1 + offscreen);
	ymin = - naxis2 * offscreen;
	ymax = naxis2 * (1 + offscreen);
	/*  Print out archive name, a complete path to the file and */
	/*  image coordinates                                       */
	if (xpix >= xmin && xpix <= xmax &&
	    ypix >= ymin && ypix <= ymax) {
	  archive = strdup(PQgetvalue(pgres, i, iarch));
	  path = strdup(PQgetvalue(pgres, i, ipath));
	  filename = strdup(PQgetvalue(pgres, i, ifile));
	  printf("%s %s/%s %7.2f %7.2f\n", archive, path, filename, xpix, ypix);
	}
      }
      else {
	err = 0;
      }
    }
  out:
  }
  PQfinish(pgconn);
  if (state.verbose) {
    printf("Connection closed!\n");
  }
  exit(0);
}



